from types import TracebackType
from typing import (Any, AsyncGenerator, Awaitable, Callable, Generator,
                    Generic, Optional, Type, TypeVar)

_S = TypeVar('_S')
_T = TypeVar('_T')
_Ca = Callable[[Generator[Awaitable, Any, None], Awaitable], AsyncGenerator[_S, _T]]

class Channel(Generic[_S, _T]):
    client: AsyncGenerator[_S, _T]
    server: AsyncGenerator[_T, _S]

    def __init__(self, cli_call: _Ca[_S, _T] = ..., ser_call: _Ca[_T, _S] = ...): ...
    def __aiter__(self) -> AsyncGenerator[_S, _T]: ...
    def __aenter__(self) -> Awaitable[AsyncGenerator[_T, _S]]: ...
    def __aexit__(self, exc_type: Optional[Type[BaseException]],
                  exc_value: Optional[BaseException],
                  traceback: Optional[TracebackType]) -> Awaitable[Optional[bool]]: ...



"""Point-to-point object streams between coroutines."""
import asyncio

__all__ = ['Channel']


def _fgen():
    """Yield futures and set sent values as their results."""
    create_future = asyncio.get_running_loop().create_future
    aitem = create_future()
    try:
        while True:
            aitem.set_result((yield aitem))
            aitem = create_future()
    finally:
        aitem.cancel()


async def _agen(gen, aitem):
    """Asynchronous generator adapter."""
    send = gen.send
    try:
        while True:
            aitem = send((yield await aitem))
    except asyncio.CancelledError:
        pass
    except GeneratorExit:
        gen.close()


class Channel:
    """Rendezvous channel between two coroutines."""

    def __init__(self, cli_call=_agen, ser_call=_agen):
        gen = _fgen()
        self.server = ser_call(gen, gen.send(None))
        self.client = cli_call(gen, gen.send(None))

    def __aiter__(self):
        return self.client

    async def __aenter__(self):
        await self.server.asend(None)
        return self.server

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.server.aclose()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sav.channels",
    version="0.3",
    author="Sander Voerman",
    author_email="sander@savoerman.nl",
    description="Iterable streams between coroutines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sandervoerman/channels",
    packages=['sav.channels'],
    package_data={
        'sav.channels': ['py.typed', '__init__.pyi'],
    },
    python_requires='>=3.7',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
